
/*
 *   OpenJupes v1.1 
 *
 *   Server and Nickname Jupe Service for Ratbox v2+services.
 *   Copyright (C) 2005 Thomas Carlsson, Joost Vunderink and Alan Milford
 *
 *   NOTE: This service depends heavily on the +services patch by Lee
 *   Hardy. It will not compile without it.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include "stdinc.h"
#include "client.h"
#include "irc_string.h"
#include "send.h"
#include "numeric.h"
#include "ircd.h"
#include "msg.h"
#include "parse.h"
#include "modules.h"
#include "s_serv.h"
#include "hash.h"
#include "s_conf.h"
#include "hook.h"
#include "services.h"

#include <time.h>
#include <string.h>

#include "oj_main.h"



#define JUPES_DB_VERSION    0x00000001


static struct OJ_jupes jupes;
static char OJ_reply[IRCLINELEN * 5];   


void OJ_Initialize()
{
    OJ_IRCD_BroadcastLocal( "OpenJupes::Initialize - Preparing memory structures." );

    jupes.nickname_activemax = JUPE_NICKJUPESMIN;
    jupes.nickname_activecount = 0;

    jupes.nickname_active = (struct OJ_nickjupe*)MyMalloc( sizeof( struct OJ_nickjupe ) * jupes.nickname_activemax );

    if ( !jupes.nickname_active )
    {
        OJ_IRCD_BroadcastLocal( "OpenJupes::Initialize - FATAL - Unable to allocate memory (machine out of memory?)" );
        jupes.nickname_activemax = 0;
    }
    else
    {
        memset( jupes.nickname_active, 0, sizeof( struct OJ_nickjupe ) * jupes.nickname_activemax );
    }

    jupes.server_activemax = JUPE_SERVERJUPESMIN;
    jupes.server_activecount = 0;

    jupes.server_active = (struct OJ_serverjupe*)MyMalloc( sizeof( struct OJ_serverjupe ) * jupes.server_activemax );

    if ( !jupes.server_active )
    {
        OJ_IRCD_BroadcastLocal( "OpenJupes::Initialize - FATAL - Unable to allocate memory (machine out of memory?)" );
        jupes.server_activemax = 0;
    }
    else
    {
        memset( jupes.server_active, 0, sizeof( struct OJ_serverjupe ) * jupes.server_activemax );
    }

    memset( jupes.server_history, 0, sizeof( struct OJ_serverjupe ) * JUPE_HISTORYMAX );
    jupes.server_historycount = 0;
}

/* static for now */
static char dbfilename[64] = "jupes.dat";

void OJ_LoadDB()
{
    int dbver, dbcrc, i;
    FILE* f;
    int data, active, pending;

    f = fopen( dbfilename, "rb" );
    if ( !f )
    {
        OJ_IRCD_BroadcastLocal( "OpenJupes::LoadDB - FATAL - unable to open file for reading." );
        return;
    }

    dbcrc = 0;
    if ( !fread( &dbver, sizeof( int ), 1, f ) ) 
    {
        fclose( f );
        OJ_IRCD_BroadcastLocal( "OpenJupes::LoadDB - FATAL - unable to open file for reading." );
        return;
    }

    if ( dbver != JUPES_DB_VERSION )
    {
        fclose( f );
        OJ_IRCD_BroadcastLocal( "OpenJupes::LoadDB - FATAL - unable to open file for reading." );
        return;
    }

    if ( !fread( &data, sizeof( int ), 1, f ) ) 
    {
        fclose( f );
        OJ_IRCD_BroadcastLocal( "OpenJupes::LoadDB - FATAL - unable to read database file." );
        return;
    }
    if ( data > 10000 || data < 0 ) /* Sanity */
    {
        fclose( f );
        OJ_IRCD_BroadcastLocal( "OpenJupes::LoadDB - FATAL - database file corrupt." );
        return;
    }
    if ( data > jupes.server_activemax )
    {
        /* Ok, current buffer isn't big enough to load them all.. Let's try to reallocate */
        struct OJ_serverjupe* m = NULL;
        int n = 0;

        n = data + JUPE_SERVERJUPESMIN;
        m = (struct OJ_serverjupe*)MyMalloc( sizeof( struct OJ_serverjupe ) * n );
        if ( !m )
        {
            fclose( f );
            OJ_IRCD_BroadcastLocal( "OpenJupes::Vote - FATAL - Unable to allocate memory (machine out of memory?)" );
            return;
        }
        memset( (char*)m, 0, sizeof( struct OJ_serverjupe ) * n );
        MyFree( (char*)jupes.server_active );

        jupes.server_active = m;
        jupes.server_activemax = n;
    }

    jupes.server_activecount = data;
    dbcrc += jupes.server_activecount;

    for ( i = 0; i < jupes.server_activecount; i++ )
    {
        if ( !fread( &jupes.server_active[i], sizeof( struct OJ_serverjupe ), 1, f ) ) 
        {
            fclose( f );
            OJ_IRCD_BroadcastLocal( "OpenJupes::LoadDB - FATAL - unable to read database file." );
            /* Safety mechanism.. We rather lose some memory here than exit with broken structs..
             * This load "should" not fail often anyway ;-)
             */
            OJ_Initialize();
            return;
        }
        jupes.server_active[i].data = NULL;
        dbcrc += jupes.server_active[i].timestamp + i;
    }

    if ( !fread( &data, sizeof( int ), 1, f ) ) 
    {
        fclose( f );
        OJ_IRCD_BroadcastLocal( "OpenJupes::LoadDB - FATAL - unable to read database file." );
        OJ_Initialize();
        return;
    }
    if ( data < 0 || data > 100000 )
    {
        fclose( f );
        OJ_IRCD_BroadcastLocal( "OpenJupes::LoadDB - FATAL - database file corrupt." );
        OJ_Initialize();
        return;
    }
    if ( data > JUPE_HISTORYMAX )
    {
        /* Could probably deal with this more cleanly, but for what it is, we'll just live with it */
        fclose( f );
        OJ_IRCD_BroadcastLocal( "OpenJupes::LoadDB - FATAL - history structure incompatible (too small to fit this database)." );
        return;
    }
    jupes.server_historycount = data;
    dbcrc += jupes.server_historycount;

    for ( i = 0; i < jupes.server_historycount; i++ )
    {
        if ( !fread( &jupes.server_history[i], sizeof( struct OJ_serverjupe ), 1, f ) ) 
        {
            fclose( f );
            OJ_IRCD_BroadcastLocal( "OpenJupes::LoadDB - FATAL - unable to read database file." );
            OJ_Initialize();
            return;
        }
        jupes.server_history[i].data = NULL;
        dbcrc += jupes.server_history[i].timestamp + i;
    }

    if ( !fread( &data, sizeof( int ), 1, f ) ) 
    {
        fclose( f );
        OJ_IRCD_BroadcastLocal( "OpenJupes::LoadDB - FATAL - unable to read database file." );
        OJ_Initialize();
        return;
    }

    fclose( f );

    if ( data != dbcrc )
    {
        OJ_IRCD_BroadcastLocal( "OpenJupes::LoadDB - FATAL - database file corrupt (CRC mismatch)." );
        OJ_Initialize();
        return;
    }

    /* Okay then, if we get this far with no problems, let's recreate the active jupe servers */
    active = 0;
    pending = 0;
    for ( i = 0; i < jupes.server_activecount; i++ )
    {
        int j, totalscore = 0;
        for ( j = 0; j < 5; j++ )
        {
            totalscore += jupes.server_active[i].projupe[j].score;
        }
        jupes.server_active[i].data = NULL;

        if ( totalscore >= 15 )
        {
            OJ_IRCD_CreateFakeServer( &jupes.server_active[i] );
            active++;
        }
        else
        {
            pending++;
        }
    }

    sprintf( OJ_reply, "Succesfully loaded database file: %d active jupes, %d pending jupe votes, %d historical records.",
        active, pending, jupes.server_historycount );
    OJ_IRCD_BroadcastLocal( OJ_reply );

    /* Ok - we have to double check that any pending votes aren't in fact expired - we'll do it in a lame
     * fashion (since this is a pretty rare issue in any case, and this function isn't performance critical..
     * .. and yes, the programmer is lazy!)
     */
    data = jupes.server_activecount;
    for ( i = 0; i < data; i++ )
    {
        OJ_ExpireVotes();
    }   
}

void OJ_CreateNickJupes( struct OJ_Conf* conf )
{
    int i;

    jupes.nickname_activecount = conf->nickcount;
    for ( i = 0; i < conf->nickcount; i++ )
    {
        strcpy( jupes.nickname_active[i].nick, conf->nick[i].nick );
        strcpy( jupes.nickname_active[i].user, conf->client.user ); /* Use Jupes client user@host for nick jupes */
        strcpy( jupes.nickname_active[i].host, conf->client.host );
        strcpy( jupes.nickname_active[i].gecos, conf->nick[i].gecos );

        OJ_IRCD_CreateFakeClient( &jupes.nickname_active[i] );
    }
}

void OJ_DeleteNickJupes( struct OJ_Conf* conf )
{
    int i;

    for ( i = 0; i < jupes.nickname_activecount; i++ )
    {
        OJ_IRCD_DeleteFakeClient( &jupes.nickname_active[i] );
    }
    jupes.nickname_activecount = 0;
}

void OJ_SaveDB()
{
    int dbver, dbcrc, i;
    FILE* f;

    f = fopen( dbfilename, "wb" );

    if ( !f )
    {
        OJ_IRCD_BroadcastLocal( "OpenJupes::SaveDB - FATAL - unable to open file for writing." );
        return;
    }

    dbcrc = 0;
    dbver = JUPES_DB_VERSION;
    if ( !fwrite( &dbver, sizeof( int ), 1, f ) ) 
    {
        fclose( f );
        OJ_IRCD_BroadcastLocal( "OpenJupes::SaveDB - FATAL - unable to open file for writing." );
        return;
    }

    /* Write active jupes */
    if ( !fwrite( &jupes.server_activecount, sizeof( int ), 1, f ) ) 
    {
        fclose( f );
        OJ_IRCD_BroadcastLocal( "OpenJupes::SaveDB - FATAL - unable to write to file (filesystem full?)" );
        return;
    }
    dbcrc += jupes.server_activecount;

    for ( i = 0; i < jupes.server_activecount; i++ )
    {
        if ( !fwrite( &jupes.server_active[i], sizeof( struct OJ_serverjupe ), 1, f ) ) 
        {
            fclose( f );
            OJ_IRCD_BroadcastLocal( "OpenJupes::SaveDB - FATAL - unable to write to file (filesystem full?)" );
            return;
        }
        dbcrc += jupes.server_active[i].timestamp + i;
    }

    /* Write historical jupes */
    if ( !fwrite( &jupes.server_historycount, sizeof( int ), 1, f ) ) 
    {
        fclose( f );
        OJ_IRCD_BroadcastLocal( "OpenJupes::SaveDB - FATAL - unable to write to file (filesystem full?)" );
        return;
    }
    dbcrc += jupes.server_historycount;

    for ( i = 0; i < jupes.server_historycount; i++ )
    {
        if ( !fwrite( &jupes.server_history[i], sizeof( struct OJ_serverjupe ), 1, f ) ) 
        {
            fclose( f );
            OJ_IRCD_BroadcastLocal( "OpenJupes::SaveDB - FATAL - unable to write to file (filesystem full?)" );
            return;
        }
        dbcrc += jupes.server_history[i].timestamp + i;
    }

    /* Write "CRC" (just a simple sanity check) */
    if ( !fwrite( &dbcrc, sizeof( int ), 1, f ) ) 
    {
        fclose( f );
        OJ_IRCD_BroadcastLocal( "OpenJupes::SaveDB - FATAL - unable to write to file (filesystem full?)" );
        return;
    }

    fclose( f );
}

void OJ_Deinitialize()
{
    int i;

    if ( jupes.nickname_active )
    {
        for ( i = 0; i < jupes.nickname_activecount; i++ )
        {
            OJ_IRCD_DeleteFakeClient( &jupes.nickname_active[i] );
        }
        MyFree( (char*)jupes.nickname_active );
        jupes.nickname_active = NULL;
    }

    if ( jupes.server_active )
    {
        for ( i = 0; i < jupes.server_activecount; i++ )
        {
            OJ_IRCD_DeleteFakeServer( &jupes.server_active[i] );
        }
        MyFree( (char*)jupes.server_active );
        jupes.server_active = NULL;
    }

    jupes.server_historycount = 0;

    OJ_IRCD_BroadcastLocal( "OpenJupes::Deinitialize - Released all memory structures." );
}

void OJ_DeleteJupeInternal( struct OJ_serverjupe* s )
{
    int i, j;

    OJ_IRCD_DeleteFakeServer( s );

    /* Add it to history */
    jupes.server_historycount++;
    if ( jupes.server_historycount > JUPE_HISTORYMAX )
    {
        jupes.server_historycount = JUPE_HISTORYMAX;
    }

    for ( i = jupes.server_historycount - 1; i > 0; i-- )
    {
        memcpy( &jupes.server_history[i], &jupes.server_history[i - 1], sizeof( struct OJ_serverjupe ) );
    }
    memcpy( &jupes.server_history[0], s, sizeof( struct OJ_serverjupe ) );
    jupes.server_history[0].expiration = time( NULL );

    /* Delete from active */
    for ( i = 0; i < jupes.server_activecount; i++ )
    {
        if ( s == &jupes.server_active[i] )
        {
            /* Move the structures around if needed */
            for( j = i; j < jupes.server_activecount - 1; j++ )
            {
                memcpy( &jupes.server_active[j], &jupes.server_active[j + 1], sizeof( struct OJ_serverjupe ) );
            }
            jupes.server_activecount--;
            return;
        }
    }

    /* We should never end up here */
    OJ_IRCD_BroadcastLocal( "OpenJupes::DeleteJupeInternal - FATAL - Delete logic failure! (BC01)" );
}

/* Here we've borrowed OpenChanfix's time_t to formatted string function O:-) */
char* OJ_FormatTime( time_t seconds )
{
#define TSLEN 64
    static char datetimestring[TSLEN];
    struct tm *stm;

    stm = localtime( &seconds );
    memset( datetimestring, 0, TSLEN );
    strftime( datetimestring, TSLEN, "%Y-%m-%d %H:%M:%S", stm );
    datetimestring[TSLEN - 1] = 0;

    return datetimestring;
}

void OJ_Vote( char* jupeserver, char* jupereason, struct OJ_voter* voter )
{
    int i, totalproscore, totalantiscore;
    struct OJ_serverjupe* s;

    if ( jupes.server_activecount >= 5 )
    {
        /*  If we have a lot of active jupes, let's do a bit of additional checking
         *   as a security precaution, in case there's something fishy going on 
         */

        int calledbyserver = 0;
        int totalscore = 0;

        for ( s = NULL, i = 0; i < jupes.server_activecount; i++ )
        {
            if ( !strcasecmp( jupes.server_active[i].server, voter->server ) )
            {
                int j;

                s = &jupes.server_active[i];

                totalscore = 0;
                for ( j = 0; j < 5; j++ )
                {
                    totalscore += s->projupe[j].score;
                }

                /* A pending jupe from voter's server? Let's increase our count */
                if ( totalscore < 15 )
                {
                    calledbyserver++;
                }
            }
        }

        if ( calledbyserver >= 5 )
        {
            /* 5 currently pending unique votes called by voter's server already? 
             * Let's ignore additional ones in case server is hacked/broken.
             */
            OJ_IRCD_BroadcastSingleOperator( voter->nick, "JUPE: Request rejected: Your server has already called 5 pending jupe votes. Please wait until these expire/trigger before you call any additional ones." );
            return;
        }
    }

    /* Let's see if the jupeserver already exists. */
    for ( s = NULL, i = 0; i < jupes.server_activecount; i++ )
    {
        if ( !strcasecmp( jupes.server_active[i].server, jupeserver ) )
        {
            s = &jupes.server_active[i];
            break;
        }
    }
    if ( !s )
    {
        /* Is someone trying to dejupe vote a nonexisting jupe? */
        if ( voter->score < 0 )
        {
            sprintf( OJ_reply, "UNJUPE: Request rejected: %s is not currently juped.", jupeserver );
            OJ_IRCD_BroadcastSingleOperator( voter->nick, OJ_reply );
            return;
        }
        if ( !jupereason || !strlen( jupereason ) )
        {
            ircsprintf( OJ_reply, "JUPE: Request rejected: To call a new vote, you must include a reason. Syntax: /msg %s JUPE <server name> <reason>", mainclient_p->client_p->name );
            OJ_IRCD_BroadcastSingleOperator( voter->nick, OJ_reply );
            return;
        }

        /* It did not exist. We'll create a new one */
        if ( jupes.server_activecount >= jupes.server_activemax )
        {
            /* Ok, we also need a bigger dynamic database */
            struct OJ_serverjupe* m = NULL;
            int n = 0;

            n = jupes.server_activemax + JUPE_SERVERJUPESMIN;
            m = (struct OJ_serverjupe*)MyMalloc( sizeof( struct OJ_serverjupe ) * n );
            if ( !m )
            {
                OJ_IRCD_BroadcastLocal( "OpenJupes::Vote - FATAL - Unable to allocate memory (machine out of memory?)" );
                return;
            }
            memset( (char*)m, 0, sizeof( struct OJ_serverjupe ) * n );
            memcpy( (char*)m, (char*)jupes.server_active, sizeof( struct OJ_serverjupe ) * jupes.server_activecount );

            MyFree( (char*)jupes.server_active );

            jupes.server_active = m;
            jupes.server_activemax = n;
        }

        /* Create new structure */
        memset( &jupes.server_active[jupes.server_activecount], 0, sizeof( struct OJ_serverjupe ) );

        strcpy( jupes.server_active[jupes.server_activecount].server, jupeserver );
        strcpy( jupes.server_active[jupes.server_activecount].reason, jupereason );

        jupes.server_active[jupes.server_activecount].timestamp = time( NULL );
        jupes.server_active[jupes.server_activecount].expiration = 0;
        jupes.server_active[jupes.server_activecount].data = NULL;

        memcpy( &jupes.server_active[jupes.server_activecount].projupe[0], voter, sizeof( struct OJ_voter ) );

        jupes.server_activecount++;

        sprintf( OJ_reply, "JUPE: %s!%s@%s (%s) has called a JUPE vote for server %s: %s",
            voter->nick, voter->user, voter->host, voter->server, jupeserver, jupereason );
        OJ_IRCD_BroadcastAllOperators( OJ_reply );
        sprintf( OJ_reply, "JUPE: If you agree with this JUPE vote: /msg %s JUPE %s",
            mainclient_p->client_p->name, jupeserver );
        OJ_IRCD_BroadcastAllOperators( OJ_reply );

	sprintf(OJ_reply, 
	    "You can check the status of this JUPE via the STATUS command.");
        OJ_IRCD_BroadcastAllOperators( OJ_reply );

        return;
    }

    /* Okay, let's do a tally here. */
    totalproscore = 0;
    totalantiscore = 0;
    for ( i = 0; i < 5; i++ )
    {
        totalproscore += s->projupe[i].score;
        totalantiscore -= s->antijupe[i].score;
    }

    if ( totalproscore >= 15 )
    {
        /* This server is already juped */
        if ( voter->score > 0 )
        {
            sprintf( OJ_reply, "JUPE: Request rejected: %s has already been juped.", jupeserver );
            OJ_IRCD_BroadcastSingleOperator( voter->nick, OJ_reply );
            return;
        }
        else
        {
            /* Deal with dejupe vote */
            if ( totalantiscore )
            {
                /* Is the server of dejupe voter already among the voters? */
                for ( i = 0; i < 5; i++ )
                {
                    if ( !strcasecmp( voter->server, s->antijupe[i].server ) )
                    {
                        /* Ho hum. */
                        sprintf( OJ_reply, "JUPE: Request rejected: An operator from %s has already voted (%s!%s@%s).", s->antijupe[i].server,
                            s->antijupe[i].nick, s->antijupe[i].user, s->antijupe[i].host );
                        OJ_IRCD_BroadcastSingleOperator( voter->nick, OJ_reply );
                        return;
                    }
                    
                    if ( !s->antijupe[i].score )
                    {
                        memcpy( &s->antijupe[i], voter, sizeof( struct OJ_voter ) );
                        sprintf( OJ_reply, "JUPE: Your vote to UNJUPE %s has been registered. To track the status of the vote: /msg %s STATUS", 
                            jupeserver, mainclient_p->client_p->name);
                        OJ_IRCD_BroadcastSingleOperator( voter->nick, OJ_reply );

                        totalantiscore -= s->antijupe[i].score;

                        if ( totalantiscore >= 15 )
                        {
                            /* Dejupe vote successful. */
                            sprintf( OJ_reply, "Server %s has been unjuped.", jupeserver );
                            OJ_IRCD_BroadcastAllOperators( OJ_reply );

                            OJ_DeleteJupeInternal( s );
                        }

                        return;
                    }
                }
            }
            else
            {
                /* Dejupe vote called */
                memcpy( &s->antijupe[0], voter, sizeof( struct OJ_voter ) );

                sprintf( OJ_reply, "UNJUPE: %s!%s@%s (%s) has called an UNJUPE vote for juped server %s.",
                    voter->nick, voter->user, voter->host, voter->server, jupeserver );
                OJ_IRCD_BroadcastAllOperators( OJ_reply );
                sprintf( OJ_reply, "UNJUPE: If you agree with this UNJUPE vote: /msg %s UNJUPE %s",
                    mainclient_p->client_p->name, jupeserver );
                OJ_IRCD_BroadcastAllOperators( OJ_reply );
            }
        }
    }
    else
    {
        /* This server has not been juped yet */
        if ( voter->score < 0 )
        {
            sprintf( OJ_reply, "UNJUPE: Request rejected: Jupe not active (vote still pending)" );
            OJ_IRCD_BroadcastSingleOperator( voter->nick, OJ_reply );
            return;
        }
        else
        {
            /* Deal with pro-jupe vote */
            if ( totalproscore )
            {
                /* Is the server of dejupe voter already among the voters? */
                for ( i = 0; i < 5; i++ )
                {
                    if ( !strcasecmp( voter->server, s->projupe[i].server ) )
                    {
                        /* Ho hum. */
                        sprintf( OJ_reply, "JUPE: Request rejected: An operator from %s has already voted (%s!%s@%s).", s->projupe[i].server,
                            s->projupe[i].nick, s->projupe[i].user, s->projupe[i].host );
                        OJ_IRCD_BroadcastSingleOperator( voter->nick, OJ_reply );
                        return;
                    }
                    
                    if ( !s->projupe[i].score )
                    {
                        memcpy( &s->projupe[i], voter, sizeof( struct OJ_voter ) );
                        sprintf( OJ_reply, "JUPE: Your vote to JUPE %s has been registered. To track the status of the vote: /msg %s STATUS", 
                            jupeserver, mainclient_p->client_p->name );
                        OJ_IRCD_BroadcastSingleOperator( voter->nick, OJ_reply );

                        totalproscore += s->projupe[i].score;

                        if ( totalproscore >= 15 )
                        {
                            /* Set timestamp to reflect JUPE start */
                            s->timestamp = s->projupe[i].timestamp;

                            /* Jupe vote successful. */
                            sprintf( OJ_reply, "Server %s has been juped.", jupeserver );
                            OJ_IRCD_BroadcastAllOperators( OJ_reply );

                            s->data = NULL;
                            OJ_IRCD_CreateFakeServer( s );
                        }

                        return;
                    }
                }
            }
            else
            {
                /* We should never end up here */
                OJ_IRCD_BroadcastLocal( "OpenJupes::Vote - FATAL - Vote logic failure! (BC03)" );
            }

        }
    }
}

/*
 *      We run OJ_ExpireVotes once per minute to kill off jupe votes which haven't been successful.
 */
void OJ_ExpireVotes()
{
    unsigned long t;
    int i, j, totalproscore, totalantiscore;
    struct OJ_serverjupe* s;

    t = time( NULL );
    for ( i = 0; i < jupes.server_activecount; i++ )
    {
        s = &jupes.server_active[i];

        /* Active jupe without quorum? */
        totalproscore = 0;
        totalantiscore = 0;
        for ( j = 0; j < 5; j++ )
        {
            totalproscore += s->projupe[j].score;
            totalantiscore -= s->antijupe[j].score;
        }

        if ( totalproscore < 15 )
        {
            /* over 15 minutes old? */
            if ( t - s->timestamp > ( 15 * 60 ) )
            {
                sprintf( OJ_reply, "Jupe vote for %s expired.", s->server );
                OJ_IRCD_BroadcastLocal( OJ_reply );

                OJ_DeleteJupeInternal( s );
                return;

                /* NOTE: In theory, if someone has flooded the JUPE service with votes,
                 * it'll take minutes to expire all. However, we mitigate this by making
                 * sure not too many jupe votes can be initiated by a single server. 
                 */
            }
        }
        else
        {
            if ( totalantiscore )
            {
                /* Calculation of unjupe vote start is a bit of a hack, but it's a ne plus ultra hack */
                if ( t - s->antijupe[0].timestamp > ( 15 * 60 ) )
                {
                    sprintf( OJ_reply, "Unjupe vote for %s expired.", s->server );
                    OJ_IRCD_BroadcastLocal( OJ_reply );

                    /* We'll just clear the unjupe variables.. i.e. unsuccessful
                     * unjupe votes won't go into the history as such. 
                     */
                    for ( j = 0; j < 5; j++ )
                    {
                        memset( &s->antijupe[j], 0, sizeof( struct OJ_voter ) );
                    }
                }
            }
        }
    }
}



/*
 *              Delivers titillating display of current activity to requester.
 *              Takes ircd-specific datapointer.
 */
void OJ_GetActive( struct IRCD_Data* id )
{
        int i, j, totalproscore, totalantiscore;
    int aj, ap, au;
        struct OJ_serverjupe* s;

        if ( !jupes.server_activecount )
        {
                /* No dice */
                sprintf( OJ_reply, "STATUS: There are currently no active or pending jupes. To see past jupe history, /msg %s HISTORY", mainclient_p->client_p->name );
                OJ_IRCD_BroadcastNative( id, OJ_reply );
                return;
        }

    /* Let's see what we have brewing in our active pot.. */
    for ( aj = 0, ap = 0, au = 0, i = 0; i < jupes.server_activecount; i++ )
    {
        s = &jupes.server_active[i];

            totalproscore = 0;
            totalantiscore = 0;
            for ( j = 0; j < 5; j++ )
            {
                    totalproscore += s->projupe[j].score;
                    totalantiscore -= s->antijupe[j].score;
            }
        if ( totalproscore < 15 )
        {
            ap++;   /* Pending */
            }
            else
            {
                aj++;   /* Jupe */
                if ( totalantiscore )
                {
                    au++;   /* Unjupe */
                }
            }

        }
        
    /* .. and tell the world */
        sprintf( OJ_reply, "STATUS: I'm keeping track of %d %s (%d active %s, %d pending JUPE %s, %d pending UNJUPE %s):",
        ( aj + ap + au ), ( ( aj + ap + au ) == 1 ) ? "event" : "events", 
        aj, ( aj == 1 ) ? "JUPE" : "JUPES",
        ap, ( ap == 1 ) ? "vote" : "votes",
        au, ( au == 1 ) ? "vote" : "votes" );

        OJ_IRCD_BroadcastNative( id, OJ_reply );
        OJ_IRCD_BroadcastNative( id, " " );

    /* Ok, let's go through the list again in more detail.. */
        for ( i = 0; i < jupes.server_activecount; i++ )
        {
                s = &jupes.server_active[i];

                totalproscore = 0;
                totalantiscore = 0;
                for ( j = 0; j < 5; j++ )
                {
                        totalproscore += s->projupe[j].score;
                        totalantiscore -= s->antijupe[j].score;
                }

                /* Active jupe without quorum? */
                if ( totalproscore < 15 )
                {
            int delta = 15 - ( ( time( NULL ) - s->timestamp ) / 60 );
            if ( delta < 0 ) delta = 0;
                        sprintf( OJ_reply, "Pending: JUPE %s (Reason: %s) -- expires in %d minutes.", 
                s->server, s->reason, delta );

                        OJ_IRCD_BroadcastNative( id, OJ_reply );
                }
                else
                {
                        sprintf( OJ_reply, "Active: JUPE %s (Reason: %s)", s->server, s->reason );
                        OJ_IRCD_BroadcastNative( id, OJ_reply );
                }

                /* Show pro-voters */
                for ( j = 0; j < 5; j++ )
                {
                        if ( s->projupe[j].score )
                        {
                                sprintf( OJ_reply, "  Vote: %s!%s@%s (%s) %s -- score %d", 
                                        s->projupe[j].nick, s->projupe[j].user, s->projupe[j].host, s->projupe[j].server,
                                        OJ_FormatTime( s->projupe[j].timestamp ), s->projupe[j].score );
                                OJ_IRCD_BroadcastNative( id, OJ_reply );
                        }
                }

                if ( totalproscore < 15 )
                {
                        sprintf( OJ_reply, "  (Vote needs %d additional points to pass)", ( 15 - totalproscore ) );
                        OJ_IRCD_BroadcastNative( id, OJ_reply );
                }
                else
                {
            if ( totalantiscore )
            {
                /* Calculating expiration of UNJUPE vote is a bit of a hack. But it's a fine hack. */
                int delta = 15 - ( ( time( NULL ) - s->antijupe[0].timestamp ) / 60 );
                if ( delta < 0 ) delta = 0;

                            sprintf( OJ_reply, "Pending: UNJUPE %s -- expires in %d minutes.", s->server, delta );
                            OJ_IRCD_BroadcastNative( id, OJ_reply );
                            for ( j = 0; j < 5; j++ )
                            {
                                    if ( s->antijupe[j].score )
                                    {
                                            sprintf( OJ_reply, "  Vote: %s!%s@%s (%s) %s -- score %d", 
                                                    s->antijupe[j].nick, s->antijupe[j].user, s->antijupe[j].host, s->antijupe[j].server,
                                                    OJ_FormatTime( s->antijupe[j].timestamp ), -s->antijupe[j].score );
                                            OJ_IRCD_BroadcastNative( id, OJ_reply );
                                    }
                            }
                            sprintf( OJ_reply, "  (Vote needs %d additional points to pass)", ( 15 - totalantiscore ) );
                            OJ_IRCD_BroadcastNative( id, OJ_reply );
            }
                }

                OJ_IRCD_BroadcastNative( id, " " );
        }
}



/*
 *      Glorious history. User asked for it, user will get it. 
 *      Takes ircd-specific datapointer, wanted # of displayed events.
 */
void OJ_GetHistory( struct IRCD_Data* id, int wantedcount )
{
    unsigned long t;
    int i, j, totalproscore, totalantiscore;
    struct OJ_serverjupe* s;

    if ( !jupes.server_historycount )
    {
        /* No dice */
        OJ_IRCD_BroadcastNative( id, "HISTORY: I have no historical jupe votes in my records. Sorry." );
        return;
    }

    if ( wantedcount > jupes.server_historycount )
    {
        sprintf( OJ_reply, "HISTORY: I have %d historical jupe votes in my database:", jupes.server_historycount );
        OJ_IRCD_BroadcastNative( id, OJ_reply );
        OJ_IRCD_BroadcastNative( id, " " );
    }
    else
    {
        sprintf( OJ_reply, "HISTORY: The past %d jupe votes (no longer active):", wantedcount );
        OJ_IRCD_BroadcastNative( id, OJ_reply );
        OJ_IRCD_BroadcastNative( id, " " );
    }

    t = time( NULL );
    for ( i = 0; i < jupes.server_historycount && i < wantedcount; i++ )
    {
        s = &jupes.server_history[i];

        totalproscore = 0;
        totalantiscore = 0;
        for ( j = 0; j < 5; j++ )
        {
            totalproscore += s->projupe[j].score;
            totalantiscore -= s->antijupe[j].score;
        }

        /* Active jupe without quorum? */
        if ( totalproscore < 15 )
        {
            sprintf( OJ_reply, "%s -- Not passed: %s (Jupe reason: %s)", OJ_FormatTime( s->timestamp ), s->server, s->reason );
            OJ_IRCD_BroadcastNative( id, OJ_reply );
        }
        else
        {
            sprintf( OJ_reply, "%s -- JUPED: %s (Jupe reason: %s)", OJ_FormatTime( s->timestamp ), s->server, s->reason );
            OJ_IRCD_BroadcastNative( id, OJ_reply );
        }

        /* Show pro-voters */
        for ( j = 0; j < 5; j++ )
        {
            if ( s->projupe[j].score )
            {
                sprintf( OJ_reply, "  Vote: %s!%s@%s (%s) %s -- score %d", 
                    s->projupe[j].nick, s->projupe[j].user, s->projupe[j].host, s->projupe[j].server,
                    OJ_FormatTime( s->projupe[j].timestamp ), s->projupe[j].score );
                OJ_IRCD_BroadcastNative( id, OJ_reply );
            }
        }

        if ( totalproscore < 15 )
        {
            sprintf( OJ_reply, "  (Vote did not get required 15 points - expired %s)", OJ_FormatTime( s->expiration ) );
            OJ_IRCD_BroadcastNative( id, OJ_reply );
        }
        else
        {
            /* A successful vote CAN NOT be among historicals unless subsequently unjuped, so.. */
            sprintf( OJ_reply, "%s -- Unjuped: %s", OJ_FormatTime( s->expiration ), s->server );
            OJ_IRCD_BroadcastNative( id, OJ_reply );
            for ( j = 0; j < 5; j++ )
            {
                if ( s->antijupe[j].score )
                {
                    sprintf( OJ_reply, "  Vote: %s!%s@%s (%s) %s -- score %d", 
                        s->antijupe[j].nick, s->antijupe[j].user, s->antijupe[j].host, s->antijupe[j].server,
                        OJ_FormatTime( s->antijupe[j].timestamp ), -s->antijupe[j].score );
                    OJ_IRCD_BroadcastNative( id, OJ_reply );
                }
            }
        }

        OJ_IRCD_BroadcastNative( id, " " );
    }
}

int OJ_IsJuped( char* jupeserver )
{
    int i;
        for ( i = 0; i < jupes.server_activecount; i++ )
        {
                if ( !strcasecmp( jupes.server_active[i].server, jupeserver ) )
                {
            return 1;
                }
        }
    return 0;
}
